@@ -2,11 +2,15 @@ from punpy.mc.mc_propagation import MCPropagation
 from punpy.jacobian.jacobian_propagation import JacobianPropagation
 from punpy.mc.MCMC_retrieval import MCMCRetrieval
 from punpy.jacobian.jacobian_retrieval import JacobianRetrieval
-
+from punpy.utilities.utilities import calculate_Jacobian, convert_corr_to_cov, convert_cov_to_corr
 from ._version import get_versions
 
-__author__ = "Sam Hunt <sam.hunt@npl.co.uk>"
-__all__ = []
+'''___Authorship___'''
+__author__ = "Pieter De Vis"
+__created__ = "01/03/2020"
+__maintainer__ = "Pieter De Vis"
+__email__ = "pieter.de.vis@npl.co.uk"
+__status__ = "Development"
 
 __version__ = get_versions()["version"]
 del get_versions
