@@ -25,7 +25,7 @@ class MCPropagation:
         if parallel_cores>1:
             self.pool=Pool(parallel_cores)
 
-    def propagate_random(self,func,x,u_x,cov_x=None,param_fixed=None,corr_between=None,return_corr=False,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1):
+    def propagate_random(self,func,x,u_x,cov_x=None,param_fixed=None,corr_between=None,return_corr=False,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1,PD_corr=True):
         """
         Propagate random uncertainties through measurement function with n input quantities.
         Input quantities can be floats, vectors (1d-array) or images (2d-array).
@@ -65,7 +65,7 @@ class MCPropagation:
             for i in range(n_repeats):
                 xb, u_xb = self.select_repeated_x(x,u_x,param_fixed,i,repeat_axis,n_repeats)
                 outs[i] = self.propagate_random(func,xb,u_xb,cov_x,param_fixed,corr_between,return_corr,
-                                                return_samples,repeat_dims,corr_axis=corr_axis,output_vars=output_vars)
+                                                return_samples,repeat_dims,corr_axis=corr_axis,output_vars=output_vars,PD_corr=False)
             return self.combine_repeated_outs(outs,yshape,n_repeats,len(x),repeat_axis,return_corr,return_samples,output_vars)
 
         else:
@@ -75,16 +75,28 @@ class MCPropagation:
                     MC_data[i] = self.generate_samples_random(x[i],u_x[i])
                 elif cov_x[i]=="syst":
                     MC_data[i] = self.generate_samples_systematic(x[i],u_x[i])
+                elif param_fixed is not None:
+                    if param_fixed[i] and (len(x[i].shape) == 2):
+                        MC_data[i] = np.array([self.generate_samples_cov(
+                                        x[i][:,j].flatten(),cov_x[i]).reshape(
+                                        x[i][:,j].shape+(self.MCsteps,)) for j in
+                                            range(x[i].shape[1])]).T
+                        MC_data[i] = np.moveaxis(MC_data[i],0,1)
+
+                    else:
+                        MC_data[i] = self.generate_samples_cov(x[i].flatten(),cov_x[i]).reshape(x[i].shape+(self.MCsteps,))
                 else:
-                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),cov_x[i]).reshape(x[i].shape+(self.MCsteps,))
+                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),
+                                                           cov_x[i]).reshape(
+                        x[i].shape+(self.MCsteps,))
 
                 #print(MC_data[i].nbytes)
             if corr_between is not None:
-                print(len(MC_data),MC_data.shape)
+                #print(len(MC_data),MC_data.shape)
                 MC_data = self.correlate_samples_corr(MC_data,corr_between)
-            return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars)
+            return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars,PD_corr)
 
-    def propagate_systematic(self,func,x,u_x,cov_x=None,param_fixed=None,corr_between=None,return_corr=False,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1):
+    def propagate_systematic(self,func,x,u_x,cov_x=None,param_fixed=None,corr_between=None,return_corr=False,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1,PD_corr=True):
         """
         Propagate systematic uncertainties through measurement function with n input quantities.
         Input quantities can be floats, vectors (1d-array) or images (2d-array).
@@ -124,10 +136,12 @@ class MCPropagation:
             outs = np.empty(n_repeats,dtype=object)
             for i in range(n_repeats):
                 xb,u_xb = self.select_repeated_x(x,u_x,param_fixed,i,repeat_axis,n_repeats)
+                # for ij in range(len(xb)):
+                #     print(ij,xb[ij],u_xb[ij])
                 outs[i] = self.propagate_systematic(func,xb,u_xb,cov_x,param_fixed,corr_between,
                                                 return_corr,return_samples,repeat_dims,
                                                 corr_axis=corr_axis,
-                                                output_vars=output_vars)
+                                                output_vars=output_vars,PD_corr=False)
             return self.combine_repeated_outs(outs,yshape,n_repeats,len(x),repeat_axis,
                                               return_corr,return_samples,output_vars)
 
@@ -136,20 +150,31 @@ class MCPropagation:
             for i in range(len(x)):
                 if u_x[i] is None:
                     u_x[i] = np.zeros_like(x[i])
-
-                if cov_x is None or cov_x[i] == "syst":
+                if cov_x[i] is None or cov_x[i] == "syst":
                     MC_data[i] = self.generate_samples_systematic(x[i],u_x[i])
                 elif cov_x[i] == "rand":
                     MC_data[i] = self.generate_samples_random(x[i],u_x[i])
+                elif param_fixed is not None:
+                    if param_fixed[i] and (len(x[i].shape) == 2):
+                        MC_data[i] = np.array([self.generate_samples_cov(
+                            x[i][:,j].flatten(),cov_x[i]).reshape
+                            (x[i][:,j].shape+(self.MCsteps,)) for j in
+                                               range(x[i].shape[1])])
+                        MC_data[i] = np.moveaxis(MC_data[i],0,1)
+                    else:
+                        MC_data[i] = self.generate_samples_cov(x[i].flatten()
+                                                               ,cov_x[i]).reshape\
+                            (x[i].shape+(self.MCsteps,))
                 else:
-                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),cov_x[i]).reshape(x[i].shape+(self.MCsteps,))
-
+                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),
+                                                           cov_x[i]).reshape(
+                        x[i].shape+(self.MCsteps,))
             if corr_between is not None:
                 MC_data = self.correlate_samples_corr(MC_data,corr_between)
 
-            return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars)
+            return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars,PD_corr)
 
-    def propagate_cov(self,func,x,cov_x,param_fixed=None,corr_between=None,return_corr=True,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1):
+    def propagate_cov(self,func,x,cov_x,param_fixed=None,corr_between=None,return_corr=True,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1,PD_corr=True):
         """
         Propagate uncertainties with given covariance matrix through measurement function with n input quantities.
         Input quantities can be floats, vectors (1d-array) or images (2d-array).
@@ -191,7 +216,7 @@ class MCPropagation:
                 outs[i] = self.propagate_cov(func,xb,cov_x,param_fixed,corr_between,
                                                 return_corr,return_samples,repeat_dims,
                                                 corr_axis=corr_axis,
-                                                output_vars=output_vars)
+                                                output_vars=output_vars,PD_corr=False)
             return self.combine_repeated_outs(outs,yshape,n_repeats,len(x),repeat_axis,
                                               return_corr,return_samples,output_vars)
         else:
@@ -201,12 +226,25 @@ class MCPropagation:
                     MC_data[i] = self.generate_samples_systematic(x[i],cov_x[i])
                 elif (all((cov_x[i]==0).flatten())): #This is the case if one of the variables has no uncertainty
                     MC_data[i] = np.tile(x[i].flatten(),(self.MCsteps,1)).T
+                elif param_fixed is not None:
+                    if param_fixed[i] and (len(x[i].shape) == 2):
+                        MC_data[i] = np.array([self.generate_samples_cov(
+                            x[i][:,j].flatten(),cov_x[i]).reshape
+                            (x[i][:,j].shape+(self.MCsteps,)) for j in
+                                               range(x[i].shape[1])]).T
+                        MC_data[i] = np.moveaxis(MC_data[i],0,1)
+                    else:
+                        MC_data[i] = self.generate_samples_cov(x[i].flatten()
+                                                               ,cov_x[i]).reshape\
+                            (x[i].shape+(self.MCsteps,))
                 else:
-                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),cov_x[i]).reshape(x[i].shape+(self.MCsteps,))
+                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),
+                                                           cov_x[i]).reshape(
+                        x[i].shape+(self.MCsteps,))
             if corr_between is not None:
                 MC_data = self.correlate_samples_corr(MC_data,corr_between)
 
-        return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars)
+        return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars,PD_corr)
 
     def perform_checks(self,func,x,u_x,repeat_dims,corr_axis,output_vars):
         if output_vars == 1:
@@ -223,7 +261,7 @@ class MCPropagation:
                 shapewarning=True
 
         if shapewarning:
-            print(
+            warnings.warn(
                 "It looks like one of your input quantities is not an array or does not the same shape as the measurand."
                 "This is not a problem, but means you likely cannot use array operations in your measurement function."
                 "You might need to set parallel_cores to 1 or higher when creating your MCPropagation object.")
@@ -257,37 +295,36 @@ class MCPropagation:
         xb=np.zeros(len(x),dtype=object)
         u_xb=np.zeros(len(u_x),dtype=object)
         for j in range(len(x)):
-            if len(x[j].shape) > repeat_axis:
-                if (x[j].shape[repeat_axis]!=n_repeats):
+            selected=False
+            if param_fixed is not None:
+                if param_fixed[j] == True:
                     xb[j] = x[j]
                     u_xb[j] = u_x[j]
-                elif param_fixed is not None:
-                    if param_fixed[j]==True:
+                    selected=True
+            if not selected:
+                if len(x[j].shape) > repeat_axis:
+                    if (x[j].shape[repeat_axis]!=n_repeats):
                         xb[j] = x[j]
                         u_xb[j] = u_x[j]
-                elif repeat_axis == 0 :
-                    xb[j]=x[j][i]
-                    u_xb[j] = u_x[j][i]
-                elif repeat_axis == 1:
-                    xb[j] = x[j][:,i]
-                    u_xb[j] = u_x[j][:,i]
-                elif repeat_axis == 2:
-                    xb[j] = x[j][:,:,i]
-                    u_xb[j] = u_x[j][:,:,i]
+                    elif repeat_axis == 0 :
+                        xb[j]=x[j][i]
+                        u_xb[j] = u_x[j][i]
+                    elif repeat_axis == 1:
+                        xb[j] = x[j][:,i]
+                        u_xb[j] = u_x[j][:,i]
+                    elif repeat_axis == 2:
+                        xb[j] = x[j][:,:,i]
+                        u_xb[j] = u_x[j][:,:,i]
+                    else:
+                        warnings.warn("The repeat axis is too large to be dealt with by the"
+                                      "current version of punpy.")
                 else:
-                    warnings.warn("The repeat axis is too large to be dealt with by the"
-                                  "current version of punpy.")
-            else:
-                if (len(x[j])==n_repeats):
-                    xb[j] = x[j][i]
-                    u_xb[j] = u_x[j][i]
-                    if param_fixed is not None:
-                        if param_fixed[j]==True:
-                            xb[j] = x[j]
-                            u_xb[j] = u_x[j]
-                else:
-                    xb[j] = x[j]
-                    u_xb[j] = u_x[j]
+                    if (len(x[j])==n_repeats):
+                        xb[j] = x[j][i]
+                        u_xb[j] = u_x[j][i]
+                    else:
+                        xb[j] = x[j]
+                        u_xb[j] = u_x[j]
         return xb, u_xb
 
     def combine_repeated_outs(self,outs,yshape,n_repeats,lenx,repeat_axis,return_corr,return_samples,output_vars):
@@ -345,11 +382,17 @@ class MCPropagation:
             returns[0]=u_func
             extra_index=0
             if return_corr:
-                returns[1]=np.mean([outs[i][1] for i in range(n_repeats)],axis=0)
+                corr=np.mean([outs[i][1] for i in range(n_repeats)],axis=0)
+                if not self.isPD(corr):
+                    corr=self.nearestPD_cholesky(corr,corr=True,return_cholesky=False)
+                returns[1]=corr
                 extra_index+=1
 
             if output_vars>1:
-                returns[1+extra_index]=np.mean([outs[i][1+extra_index] for i in range(n_repeats)],axis=0)
+                corr_out = np.mean([outs[i][1+extra_index] for i in range(n_repeats)],axis=0)
+                if not self.isPD(corr_out):
+                    corr_out=self.nearestPD_cholesky(corr_out,corr=True,return_cholesky=False)
+                returns[1+extra_index]=corr_out
                 extra_index+=1
 
             if return_samples:
@@ -358,7 +401,7 @@ class MCPropagation:
 
             return returns
 
-    def process_samples(self,func,data,return_corr,return_samples,corr_axis=-99,output_vars=1):
+    def process_samples(self,func,data,return_corr,return_samples,corr_axis=-99,output_vars=1,PD_corr=True):
         """
         Run the MC-generated samples of input quantities through the measurement function and calculate
         correlation matrix if required.
@@ -398,6 +441,7 @@ class MCPropagation:
             MC_y = np.moveaxis(MC_y2,0,-1)
 
         u_func = np.std(MC_y,axis=-1)
+
         if not return_corr:
             if return_samples:
                 return u_func,MC_y,data
@@ -405,7 +449,11 @@ class MCPropagation:
                 return u_func
         else:
             if output_vars==1:
-                corr_y = self.calculate_corr(MC_y,corr_axis)
+                corr_y = self.calculate_corr(MC_y,corr_axis).astype("float32")
+                if PD_corr:
+                    if not self.isPD(corr_y):
+                        corr_y = self.nearestPD_cholesky(corr_y,corr=True,
+                                                       return_cholesky=False)
                 if return_samples:
                     return u_func,corr_y,MC_y,data
                 else:
@@ -415,10 +463,17 @@ class MCPropagation:
                 #create an empty arrays and then populate it with the correlation matrix for each output parameter individually
                 corr_ys=np.empty(output_vars,dtype=object)
                 for i in range(output_vars):
-                    corr_ys[i] = self.calculate_corr(MC_y[i],corr_axis)
-
+                    corr_ys[i] = self.calculate_corr(MC_y[i],corr_axis).astype("float32")
+                    if PD_corr:
+                        if not self.isPD(corr_ys[i]):
+                            corr_ys[i] = self.nearestPD_cholesky(corr_ys[i],corr=True,
+                                                             return_cholesky=False)
                 #calculate correlation matrix between the different outputs produced by the measurement function.
-                corr_out=np.corrcoef(MC_y.reshape((output_vars,-1)))
+                corr_out=np.corrcoef(MC_y.reshape((output_vars,-1))).astype("float32")
+                if PD_corr:
+                    if not self.isPD(corr_out):
+                        corr_out = self.nearestPD_cholesky(corr_out,corr=True,
+                                                         return_cholesky=False)
 
                 if return_samples:
                     return u_func,corr_ys,corr_out,MC_y,data
@@ -592,7 +647,7 @@ class MCPropagation:
                 return samples_out
 
     @staticmethod
-    def nearestPD_cholesky(A):
+    def nearestPD_cholesky(A,diff=0.001,corr=False,return_cholesky=True):
         """
         Find the nearest positive-definite matrix
 
@@ -632,14 +687,38 @@ class MCPropagation:
                 A3 += I*(-mineig*k**2+spacing)
                 k += 1
 
-            if np.any(abs(A-A3)/(A+0.0001) > 0.0001):
-                raise ValueError(
-                    "One of the provided covariance matrix is not postive definite. Covariance matrices need to be at least positive semi-definite. Please check your covariance matrix.")
-            else:
-                warnings.warn(
-                    "One of the provided covariance matrix is not positive definite. It has been slightly changed (less than 0.01% in any element) to accomodate our method.")
-                return np.linalg.cholesky(A3)
+            maxdiff=np.max(np.abs(A-A3))
 
+            if corr == True:
+                if maxdiff>diff:
+                    raise ValueError(
+                        "One of the correlation matrices is not postive definite. "
+                        "Correlation matrices need to be at least positive "
+                        "semi-definite.")
+                else:
+                    print(
+                        "One of the provided covariance matrix is not positive "
+                        "definite. It has been slightly changed (less than %s in any "
+                        "element) to accomodate our method."%(diff))
+                    if return_cholesky:
+                        return np.linalg.cholesky(A3)
+                    else:
+                        return A3
+            else:
+                if maxdiff > diff:
+                    raise ValueError(
+                        "One of the provided covariance matrices is not postive "
+                        "definite. Covariance matrices need to be at least positive "
+                        "semi-definite. Please check your covariance matrix.")
+                else:
+                    print(
+                        "One of the provided covariance matrix is not positive "
+                        "definite. It has been slightly changed (less than %s \% in "
+                        "any element) to accomodate our method."%(diff/100))
+                    if return_cholesky:
+                        return np.linalg.cholesky(A3)
+                    else:
+                        return A3
     @staticmethod
     def isPD(B):
         """
